{% extends "auth/page.html.j2" %}

{% block login_widget %}
{% endblock %}

{% block main %}
<div class="container col-md-6 col-md-offset-3">
    <h1 class="text-center">Authorize access</h1>

    <h2>
        An application is requesting authorization to access data associated with your Grader Service account
    </h2>

    <p>
        {{ oauth_client.description }} (oauth URL: {{ oauth_client.redirect_uri }})
        would like permission to identify you.
    </p>

    <h3>This will grant the application permission to:</h3>
    <div>
        <form method="POST" action="">
            <input type="hidden" name="_xsrf" value="{{ xsrf }}"/>
            <input type="submit" value="Authorize" class="form-control btn-jupyter"/>
        </form>
    </div>
</div>

{% endblock %}
