from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class FeedbackStatus(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    """
    allowed enum values
    """
    NOT_GENERATED = 'not_generated'
    GENERATING = 'generating'
    GENERATED = 'generated'
    GENERATION_FAILED = 'generation_failed'
    FEEDBACK_OUTDATED = 'feedback_outdated'
    def __init__(self):  # noqa: E501
        """FeedbackStatus - a model defined in OpenAPI

        """
        self.openapi_types = {
        }

        self.attribute_map = {
        }

    @classmethod
    def from_dict(cls, dikt) -> 'FeedbackStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The FeedbackStatus of this FeedbackStatus.  # noqa: E501
        :rtype: FeedbackStatus
        """
        return util.deserialize_model(dikt, cls)
