from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from grader_service.api.models.base_model import Model
from grader_service.api import util


class ManualStatus(Model):
    """NOTE: This class is auto generated by OpenAPI Generator (https://openapi-generator.tech).

    Do not edit the class manually.
    """

    """
    allowed enum values
    """
    NOT_GRADED = 'not_graded'
    MANUALLY_GRADED = 'manually_graded'
    BEING_EDITED = 'being_edited'
    def __init__(self):  # noqa: E501
        """ManualStatus - a model defined in OpenAPI

        """
        self.openapi_types = {
        }

        self.attribute_map = {
        }

    @classmethod
    def from_dict(cls, dikt) -> 'ManualStatus':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The ManualStatus of this ManualStatus.  # noqa: E501
        :rtype: ManualStatus
        """
        return util.deserialize_model(dikt, cls)
